@@ -1,8 +1,13 @@
 use crate::block::{ItemContent, ItemPosition, Prelim};
-use crate::types::{Branch, BranchRef, Entries, Event, Observer, TypePtr, Value, TYPE_REFS_MAP};
+use crate::event::Subscription;
+use crate::types::{
+    event_keys, Branch, BranchRef, Entries, EntryChange, Observers, Path, TypePtr, Value,
+    TYPE_REFS_MAP,
+};
 use crate::*;
 use lib0::any::Any;
-use std::collections::HashMap;
+use std::cell::UnsafeCell;
+use std::collections::{HashMap, HashSet};
 use std::rc::Rc;
 
 /// Collection used to store key-value entries in an unordered manner. Keys are always represented
@@ -91,6 +96,7 @@ impl Map {
                 left: left.cloned(),
                 right: None,
                 index: 0,
+                current_attrs: None,
             }
         };
 
@@ -138,12 +144,24 @@ impl Map {
     /// All map changes can be tracked by using [Event::keys] method.
     ///
     /// Returns an [Observer] which, when dropped, will unsubscribe current callback.
-    pub fn observe<F>(&self, f: F) -> Observer
+    pub fn observe<F>(&self, f: F) -> Subscription<MapEvent>
     where
-        F: Fn(&Transaction, &Event) -> () + 'static,
+        F: Fn(&Transaction, &MapEvent) -> () + 'static,
     {
         let mut branch = self.0.borrow_mut();
-        branch.observe(f)
+        if let Observers::Map(eh) = branch.observers.get_or_insert_with(Observers::map) {
+            eh.subscribe(f)
+        } else {
+            panic!("Observed collection is of different type") //TODO: this should be Result::Err
+        }
+    }
+
+    /// Unsubscribes a previously subscribed event callback identified by given `subscription_id`.
+    pub fn unobserve(&self, subscription_id: SubscriptionId) {
+        let mut branch = self.0.borrow_mut();
+        if let Some(Observers::Map(eh)) = branch.observers.as_mut() {
+            eh.unsubscribe(subscription_id);
+        }
     }
 }
 
@@ -216,6 +234,55 @@ impl<T: Prelim> Prelim for PrelimMap<T> {
     }
 }
 
+/// Event generated by [Map::observe] method. Emitted during transaction commit phase.
+pub struct MapEvent {
+    target: Map,
+    current_target: BranchRef,
+    keys: UnsafeCell<Result<HashMap<Rc<str>, EntryChange>, HashSet<Option<Rc<str>>>>>,
+}
+
+impl MapEvent {
+    pub(crate) fn new(branch_ref: BranchRef, key_changes: HashSet<Option<Rc<str>>>) -> Self {
+        let current_target = branch_ref.clone();
+        MapEvent {
+            target: Map::from(branch_ref),
+            current_target,
+            keys: UnsafeCell::new(Err(key_changes)),
+        }
+    }
+
+    /// Returns a [Map] instance which emitted this event.
+    pub fn target(&self) -> &Map {
+        &self.target
+    }
+
+    /// Returns a path from root type down to [Map] instance which emitted this event.
+    pub fn path(&self, txn: &Transaction) -> Path {
+        Branch::path(self.current_target.borrow(), self.target.0.borrow(), txn)
+    }
+
+    /// Returns a summary of key-value changes made over corresponding [Map] collection within
+    /// bounds of current transaction.
+    pub fn keys(&self, txn: &Transaction) -> &HashMap<Rc<str>, EntryChange> {
+        let keys = unsafe { self.keys.get().as_mut().unwrap() };
+
+        match keys {
+            Ok(keys) => {
+                return keys;
+            }
+            Err(subs) => {
+                let subs = event_keys(txn, self.target.0.borrow(), subs);
+                *keys = Ok(subs);
+                if let Ok(keys) = keys {
+                    keys
+                } else {
+                    panic!("Defect: should not happen");
+                }
+            }
+        }
+    }
+}
+
 #[cfg(test)]
 mod test {
     use crate::test_utils::{exchange_updates, run_scenario};
